@@ -37,6 +37,13 @@
 #include <gnutls/crypto.h>
 #include <time.h>
 #include <common.h>
+#include <unistd.h>
+
+#ifndef _WIN32
+# include <signal.h>
+#else
+#include <ws2tcpip.h>
+#endif
 
 #ifdef ENABLE_PKCS11
 #include <gnutls/pkcs11.h>
@@ -46,7 +53,7 @@
 
 const char str_unknown[] = "(unknown)";
 
-/* Hex encodes the given data.
+/* Hex encodes the given data adding a semicolon between hex bytes.
  */
 const char *raw_to_string(const unsigned char *raw, size_t raw_size)
 {
@@ -67,6 +74,46 @@ const char *raw_to_string(const unsigned char *raw, size_t raw_size)
 	return buf;
 }
 
+/* Hex encodes the given data.
+ */
+const char *raw_to_hex(const unsigned char *raw, size_t raw_size)
+{
+	static char buf[1024];
+	size_t i;
+	if (raw_size == 0)
+		return "(empty)";
+
+	if (raw_size * 2 + 1 >= sizeof(buf))
+		return "(too large)";
+
+	for (i = 0; i < raw_size; i++) {
+		sprintf(&(buf[i * 2]), "%02x", raw[i]);
+	}
+	buf[sizeof(buf) - 1] = '\0';
+
+	return buf;
+}
+
+const char *raw_to_base64(const unsigned char *raw, size_t raw_size)
+{
+	static char buf[1024];
+	gnutls_datum_t data = {(unsigned char*)raw, raw_size};
+	size_t buf_size;
+	int ret;
+
+	if (raw_size == 0)
+		return "(empty)";
+
+	buf_size = sizeof(buf);
+	ret = gnutls_pem_base64_encode(NULL, &data, buf, &buf_size);
+	if (ret < 0)
+		return "(error)";
+
+	buf[sizeof(buf) - 1] = '\0';
+
+	return buf;
+}
+
 static void print_x509_info_compact(gnutls_session_t session)
 {
 	gnutls_x509_crt_t crt;
@@ -101,7 +148,7 @@ static void print_x509_info_compact(gnutls_session_t session)
 }
 
 static void
-print_x509_info(gnutls_session_t session, int flag, int print_cert)
+print_x509_info(gnutls_session_t session, FILE *out, int flag, int print_cert)
 {
 	gnutls_x509_crt_t crt;
 	const gnutls_datum_t *cert_list;
@@ -114,8 +161,8 @@ print_x509_info(gnutls_session_t session, int flag, int print_cert)
 		return;
 	}
 
-	printf("- Certificate type: X.509\n");
-	printf("- Got a certificate list of %d certificates.\n",
+	fprintf(out, "- Certificate type: X.509\n");
+	fprintf(out, "- Got a certificate list of %d certificates.\n",
 	       cert_list_size);
 
 	for (j = 0; j < cert_list_size; j++) {
@@ -131,13 +178,13 @@ print_x509_info(gnutls_session_t session, int flag, int print_cert)
 			return;
 		}
 
-		printf("- Certificate[%d] info:\n - ", j);
+		fprintf(out, "- Certificate[%d] info:\n - ", j);
 		if (flag == GNUTLS_CRT_PRINT_COMPACT && j > 0)
 			flag = GNUTLS_CRT_PRINT_ONELINE;
 
 		ret = gnutls_x509_crt_print(crt, flag, &cinfo);
 		if (ret == 0) {
-			printf("%s\n", cinfo.data);
+			fprintf(out, "%s\n", cinfo.data);
 			gnutls_free(cinfo.data);
 		}
 
@@ -168,9 +215,9 @@ print_x509_info(gnutls_session_t session, int flag, int print_cert)
 			}
 
 			p[size] = 0;
-			fputs("\n", stdout);
-			fputs(p, stdout);
-			fputs("\n", stdout);
+			fputs("\n", out);
+			fputs(p, out);
+			fputs("\n", out);
 
 			gnutls_free(p);
 		}
@@ -215,7 +262,7 @@ static void print_openpgp_info_compact(gnutls_session_t session)
 }
 
 static void
-print_openpgp_info(gnutls_session_t session, int flag, int print_cert)
+print_openpgp_info(gnutls_session_t session, FILE *out, int flag, int print_cert)
 {
 
 	gnutls_openpgp_crt_t crt;
@@ -223,7 +270,7 @@ print_openpgp_info(gnutls_session_t session, int flag, int print_cert)
 	unsigned int cert_list_size = 0;
 	int ret;
 
-	printf("- Certificate type: OpenPGP\n");
+	fprintf(out, "- Certificate type: OpenPGP\n");
 
 	cert_list = gnutls_certificate_get_peers(session, &cert_list_size);
 
@@ -241,7 +288,7 @@ print_openpgp_info(gnutls_session_t session, int flag, int print_cert)
 
 		ret = gnutls_openpgp_crt_print(crt, flag, &cinfo);
 		if (ret == 0) {
-			printf("- %s\n", cinfo.data);
+			fprintf(out, "- %s\n", cinfo.data);
 			gnutls_free(cinfo.data);
 		}
 
@@ -271,8 +318,8 @@ print_openpgp_info(gnutls_session_t session, int flag, int print_cert)
 				return;
 			}
 
-			fputs(p, stdout);
-			fputs("\n", stdout);
+			fputs(p, out);
+			fputs("\n", out);
 
 			gnutls_free(p);
 		}
@@ -285,14 +332,30 @@ print_openpgp_info(gnutls_session_t session, int flag, int print_cert)
 
 /* returns false (0) if not verified, or true (1) otherwise 
  */
-int cert_verify(gnutls_session_t session, const char *hostname)
+int cert_verify(gnutls_session_t session, const char *hostname, const char *purpose)
 {
 	int rc;
 	unsigned int status = 0;
 	gnutls_datum_t out;
 	int type;
+	gnutls_typed_vdata_st data[2];
+	unsigned elements = 0;
 
-	rc = gnutls_certificate_verify_peers3(session, hostname, &status);
+	memset(data, 0, sizeof(data));
+
+	if (hostname) {
+		data[elements].type = GNUTLS_DT_DNS_HOSTNAME;
+		data[elements].data = (void*)hostname;
+		elements++;
+	}
+
+	if (purpose) {
+		data[elements].type = GNUTLS_DT_KEY_PURPOSE_OID;
+		data[elements].data = (void*)purpose;
+		elements++;
+	}
+
+	rc = gnutls_certificate_verify_peers(session, data, elements, &status);
 	if (rc == GNUTLS_E_NO_CERTIFICATE_FOUND) {
 		printf("- Peer did not send any certificate.\n");
 		return 0;
@@ -326,6 +389,7 @@ int cert_verify(gnutls_session_t session, const char *hostname)
 static void
 print_dh_info(gnutls_session_t session, const char *str, int print)
 {
+#if defined(ENABLE_DHE) || defined(ENABLE_ANON)
 	printf("- %sDiffie-Hellman parameters\n", str);
 	printf(" - Using prime: %d bits\n",
 	       gnutls_dh_get_prime_bits(session));
@@ -399,6 +463,7 @@ print_dh_info(gnutls_session_t session, const char *str, int print)
 		gnutls_free(raw_gen.data);
 		gnutls_dh_params_deinit(dh_params);
 	}
+#endif
 }
 
 static void print_ecdh_info(gnutls_session_t session, const char *str)
@@ -415,7 +480,7 @@ static void print_ecdh_info(gnutls_session_t session, const char *str)
 
 }
 
-int print_info(gnutls_session_t session, int verbose, int print_cert)
+int print_info(gnutls_session_t session, int verbose, int flags)
 {
 	const char *tmp;
 	gnutls_credentials_type_t cred;
@@ -495,8 +560,11 @@ int print_info(gnutls_session_t session, int verbose, int print_cert)
 			}
 		}
 
-		if (print_cert)
-			print_cert_info(session, verbose, print_cert);
+		if ((flags & P_WAIT_FOR_CERT) && gnutls_certificate_get_ours(session) == 0)
+			printf("- No certificate was sent to peer\n");
+
+		if (flags& P_PRINT_CERT)
+			print_cert_info(session, verbose, (flags&P_PRINT_CERT));
 
 		if (kx == GNUTLS_KX_DHE_RSA || kx == GNUTLS_KX_DHE_DSS)
 			print_dh_info(session, "Ephemeral ", verbose);
@@ -539,6 +607,20 @@ int print_info(gnutls_session_t session, int verbose, int print_cert)
 	       (gnutls_compression_get(session)));
 	printf("- Compression: %s\n", tmp);
 
+	printf("- Options:");
+	if (gnutls_session_ext_master_secret_status(session)!=0)
+		printf(" extended master secret,");
+	if (gnutls_safe_renegotiation_status(session)!=0)
+		printf(" safe renegotiation,");
+	if (gnutls_session_etm_status(session)!=0)
+		printf(" EtM,");
+#ifdef ENABLE_OCSP
+	if (gnutls_ocsp_status_request_is_checked(session, GNUTLS_OCSP_SR_IS_AVAIL)!=0) {
+		printf(" OCSP status request%s,", gnutls_ocsp_status_request_is_checked(session,0)!=0?"":"[ignored]");
+	}
+#endif
+	printf("\n");
+
 #ifdef ENABLE_DTLS_SRTP
 	rc = gnutls_srtp_get_selected_profile(session, &srtp_profile);
 	if (rc == 0)
@@ -581,6 +663,11 @@ int print_info(gnutls_session_t session, int verbose, int print_cert)
 }
 
 void print_cert_info(gnutls_session_t session, int verbose, int print_cert)
+{
+	print_cert_info2(session, verbose, stdout, print_cert);
+}
+
+void print_cert_info2(gnutls_session_t session, int verbose, FILE *out, int print_cert)
 {
 	int flag;
 
@@ -594,11 +681,11 @@ void print_cert_info(gnutls_session_t session, int verbose, int print_cert)
 
 	switch (gnutls_certificate_type_get(session)) {
 	case GNUTLS_CRT_X509:
-		print_x509_info(session, flag, print_cert);
+		print_x509_info(session, out, flag, print_cert);
 		break;
 #ifdef ENABLE_OPENPGP
 	case GNUTLS_CRT_OPENPGP:
-		print_openpgp_info(session, flag, print_cert);
+		print_openpgp_info(session, out, flag, print_cert);
 		break;
 #endif
 	default:
@@ -962,6 +1049,7 @@ pin_callback(void *user, int attempt, const char *token_url,
 	     size_t pin_max)
 {
 	const char *password = NULL;
+	common_info_st *info = user;
 	const char *desc;
 	int cache = MAX_CACHE_TRIES;
 	unsigned len;
@@ -973,9 +1061,13 @@ pin_callback(void *user, int attempt, const char *token_url,
 	if (flags & GNUTLS_PIN_SO) {
 		env = "GNUTLS_SO_PIN";
 		desc = "security officer";
+		if (info)
+			password = info->so_pin;
 	} else {
 		env = "GNUTLS_PIN";
 		desc = "user";
+		if (info)
+			password = info->pin;
 	}
 
 	if (flags & GNUTLS_PIN_FINAL_TRY) {
@@ -995,7 +1087,7 @@ pin_callback(void *user, int attempt, const char *token_url,
 	if (cache > 0 && cached_url != NULL) {
 		if (token_url != NULL
 		    && strcmp(cached_url, token_url) == 0) {
-			if (strlen(pin) >= sizeof(cached_pin)) {
+			if (strlen(cached_pin) >= pin_max) {
 				fprintf(stderr, "Too long PIN given\n");
 				exit(1);
 			}
@@ -1009,24 +1101,31 @@ pin_callback(void *user, int attempt, const char *token_url,
 		}
 	}
 
-	printf("Token '%s' with URL '%s' ", token_label, token_url);
-	printf("requires %s PIN\n", desc);
-	
-	password = getenv(env);
-	if (env == NULL) /* compatibility */
-		password = getenv("GNUTLS_PIN");
-
 	if (password == NULL) {
+		password = getenv(env);
+		if (password == NULL) /* compatibility */
+			password = getenv("GNUTLS_PIN");
+	}
+
+	if (password == NULL && (info == NULL || info->batch == 0)) {
+		fprintf(stderr, "Token '%s' with URL '%s' ", token_label, token_url);
+		fprintf(stderr, "requires %s PIN\n", desc);
+
 		password = getpass("Enter PIN: ");
 	} else {
 		if (flags & GNUTLS_PIN_WRONG) {
+			fprintf(stderr, "Token '%s' with URL '%s' ", token_label, token_url);
+			fprintf(stderr, "requires %s PIN\n", desc);
 			fprintf(stderr, "Cannot continue with a wrong password in the environment.\n");
 			exit(1);
 		}
 	}
 
 	if (password == NULL || password[0] == 0 || password[0] == '\n') {
-		fprintf(stderr, "No PIN given\n");
+		fprintf(stderr, "No PIN given.\n");
+		if (info != NULL && info->batch != 0) {
+			fprintf(stderr, "note: when operating in batch mode, set the GNUTLS_PIN or GNUTLS_SO_PIN environment variables\n");
+		}
 		exit(1);
 	}
 
@@ -1058,8 +1157,9 @@ static int
 token_callback(void *user, const char *label, const unsigned retry)
 {
 	char buf[32];
+	common_info_st *info = user;
 
-	if (retry > 0) {
+	if (retry > 0 || (info != NULL && info->batch != 0)) {
 		fprintf(stderr, "Could not find token %s\n", label);
 		return -1;
 	}
@@ -1070,12 +1170,27 @@ token_callback(void *user, const char *label, const unsigned retry)
 	return 0;
 }
 
-void pkcs11_common(void)
+void pkcs11_common(common_info_st *c)
 {
 
-	gnutls_pkcs11_set_pin_function(pin_callback, NULL);
-	gnutls_pkcs11_set_token_function(token_callback, NULL);
+	gnutls_pkcs11_set_pin_function(pin_callback, c);
+	gnutls_pkcs11_set_token_function(token_callback, c);
 
 }
 
 #endif
+
+void sockets_init(void)
+{
+#ifdef _WIN32
+	WORD wVersionRequested;
+	WSADATA wsaData;
+
+	wVersionRequested = MAKEWORD(1, 1);
+	if (WSAStartup(wVersionRequested, &wsaData) != 0) {
+		perror("WSA_STARTUP_ERROR");
+	}
+#else
+	signal(SIGPIPE, SIG_IGN);
+#endif
+}
